#include "stdio.h"
#include "stdlib.h"
#include "stddef.h"
#include "stdint.h"
#include "string.h"

#define gamma our_gamma		/* namespace problem with libc function */

#define ifnot(x) if (!(x))
#define nobreak

// ANSI color escape seq macros, all same width before expansion
#define RED "\x1b[1;31m"
#define GRE "\x1b[0;32m"
#define YEL "\x1b[1;33m"
#define CYA "\x1b[1;34m"
#define PUR "\x1b[1;35m"
#define TEA "\x1b[1;36m"
#define WHI "\x1b[1;37m"
#define BAW "\x1b[m"

typedef unsigned int  uint;
typedef unsigned long ulong;

/*
 *	TODO these are obsolete
 */
enum e_int {
	c2 = 2, c5 = 5, c6 = 6, c10 = 10, c12 = 12, c16 = 16, c18 = 18, c24 = 24,
	c30 = 30, c32 = 32, c36 = 36, c63 = 63, c64 = 64, c127 = 127, c128 = 128,
};

// For simple on/off bit setting, etc.
enum {
	bit0 = 1, bit1 = 2, bit2 = 4, bit3 = 8, bit4 = 16, bit5 = 32, bit6 = 64,
	bit7 = 128, bit8 = 256, bit9 = 512, bit10 = 1024, bit11 = 2048,
	bit12 = 4096, bit13 = 8192, bit14 = 16384, bit15 = 32768, bit16 = 65536,
	bit17 = 131072,
};

/* TODO enums in this simulation suffer from naming, semantics,
   legibility, and reliability. Enumerate mitigation approaches. */
enum alpha5 {
	a5_cmp = 0,
	a5_add = 1,
	a5_sub = 2,
	a5_rsub = 3,
	a5_casr = 3,		/* mask to isolate cmp/add/sub/rsub */

	a5_r_signed = 4,
	a5_l_signed = 8,
	a5_yes = 16,

	a5uu = a5_yes,
	a5us = a5_yes | a5_r_signed,
	a5su = a5_yes | a5_l_signed,
	a5ss = a5_yes | a5_l_signed | a5_r_signed,
};

/*
 * These are "real" opcodes, so pseudo-opcodes like INC and DEC
 * do not make the list. Multiplicities are needed for operations
 * that have specific signednesses.
 *
 * Order of signedness tags: left operand, right operand, result.
 */
enum opcodes {
	oc_A_uuu,
	oc_A_uus,
	oc_A_usu,
	oc_A_uss,
	oc_A_suu,
	oc_A_sus,
	oc_A_ssu,
	oc_A_sss,
	oc_S_uuu,
	oc_S_uus,
	oc_S_usu,
	oc_S_uss,
	oc_S_suu,
	oc_S_sus,
	oc_S_ssu,
	oc_S_sss,
	oc_RS_uuu,
	oc_RS_uus,
	oc_RS_usu,
	oc_RS_uss,
	oc_RS_suu,
	oc_RS_sus,
	oc_RS_ssu,
	oc_RS_sss,
	oc_AC_uuu,
	oc_AC_uus,
	oc_AC_usu,
	oc_AC_uss,
	oc_AC_suu,
	oc_AC_sus,
	oc_AC_ssu,
	oc_AC_sss,
	oc_SC_uuu,
	oc_SC_uus,
	oc_SC_usu,
	oc_SC_uss,
	oc_SC_suu,
	oc_SC_sus,
	oc_SC_ssu,
	oc_SC_sss,
	oc_RSC_uuu,
	oc_RSC_uus,
	oc_RSC_usu,
	oc_RSC_uss,
	oc_RSC_suu,
	oc_RSC_sus,
	oc_RSC_ssu,
	oc_RSC_sss,
	oc_IGF,
	oc_IGT,
	oc_XL,
	oc_XR,
	oc_NL,
	oc_NR,
	oc_AND,
	oc_NAND,
	oc_OR,
	oc_NOR,
	oc_XOR,
	oc_XNOR,
	oc_LANR,
	oc_RANL,
	oc_LONR,
	oc_RONL,
	oc_MAX_uuu,
	oc_MAX_uus,
	oc_MAX_usu,
	oc_MAX_uss,
	oc_MAX_suu,
	oc_MAX_sus,
	oc_MAX_ssu,
	oc_MAX_sss,
	oc_MIN_uuu,
	oc_MIN_uus,
	oc_MIN_usu,
	oc_MIN_uss,
	oc_MIN_suu,
	oc_MIN_sus,
	oc_MIN_ssu,
	oc_MIN_sss,
	oc_BOUND_u,
	oc_BOUND_s,
	oc_CMP_uu,
	oc_CMP_us,
	oc_CMP_su,
	oc_CMP_ss,
	oc_ASL_uu,
	oc_ASL_us,
	oc_ASL_su,
	oc_ASL_ss,
	oc_ASR_uu,
	oc_ASR_us,
	oc_ASR_su,
	oc_ASR_ss,
	oc_LSL,
	oc_LSR,
	oc_ROL,
	oc_ML,
	oc_MH,
	oc_MHNS,
	oc_DSL,
	oc_COPY,
	oc_NUDGE,
	oc_CRF,
	oc_PIT,
	oc_PAT,
	oc_PAIT,
	oc_MIX,
	oc_XIM,
	oc_NO_MORE,
};

enum t_vision {
	/* use powers of 2 */
	t_enable = 1,			/* pass T flag to theta (else pass 1) */
	t_not_inverted = 2,		/* invert T (or 1) if this is zero */

	/* programmer-friendly option names */
	/* carry to ALU = (t_flag NAND t_enable) XOR t_not_inverted */
	/* 'n'ormal, 'i'nverted, '0' force, '1' force */
	t_n = t_enable | t_not_inverted,
	t_i = t_enable,
	t_0 = t_not_inverted,
	t_1 = 0,				/* hideous to contemplate; see WH 1.30 */
};

enum flag_scheme {
	fs_logic,				// non-numeric: Z and N come from gamma
	fs_cmp,					// non-overflow: N comes from alpha
	fs_add_s,
	fs_add_u,
	fs_bound,
	fs_asl_uu,
	fs_asl_us,
	fs_asl_su,
	fs_asl_ss,
	fs_asr_uu,
	fs_asr_us,
	fs_asr_su,
	fs_asr_ss,
	fs_clear_r,
	fs_unallocated_1,
	fs_unallocated_2,
};

/*
 * Control decoder implementation of said opcodes.
 * Four mappings and one validation must be done:
 * the mappings must convert operation names to slot numbers,
 * and the validation must verify that opcodes are at their
 * correct array indices. XXX maybe move to decoder.c
 */
struct decoder {
	int opcode;
	char *name;
	int alpha;		/* bits: a3 a2 a1 a0 */
	int beta;
	int gamma;
	int theta;
	int alpha5;		/* replaces only bits: a4 a3 a2 */
	int t_vision;	/* ALU's vision of incoming T flag */
	int flag_scheme;
};

/* alpha.c */
void setup_alpha(void);
long alpha_sim(int oper8n, int op5, long l, long r,
	int *props, int *carries, int *t_mux, int *n_mux);
uint16_t alpha_calc(int subword, int i);

enum {
	a_carry_bit = 1 << 6,
	a_prop_bit = 1 << 7,
};

/* alu.c */
long alu(int opcode, long l, long r, int *flags);
ulong wire(ulong dst, int d, ulong src, int s, int wi);
void setup_alu(void);

/* beta.c */
void setup_beta(void);
long beta_sim(int operation, long l, long r);
uint16_t beta_calc(int subword, int i);

/* decoder.c */
void setup_decoder(void);

extern char *alpha_slots;
extern char * beta_slots;
extern char *gamma_slots;
extern char *theta_slots;

extern struct decoder Decoder[];

extern uint *DecoderRAM;

/* flags.c */
uint16_t flags_calc(int i);

/* gamma.c */
void setup_gamma(void);
long gamma_sim(int operation, int carries, long l, long r, int *zeros);
uint16_t gamma_calc(int subword, int i);

enum { g_zero_bit = 1 << 6 };

/* perms.c */
struct permutation_test_case {
	long pit1, pit2;					// 6 requisite instructions
	long pat3, pat4;
	long pit5, pit6;

	long left_operand;					// 36-bit rand from test.c
	long expected_output;				// 36-bit result we believe is correct

	int watch;							// set true for pretty displays
};

int permute(int perm, int i);
void test_perms(void);
void permutation_test_vector(struct permutation_test_case *ptc);

/* sbox.c */
extern char *alpha_sbox[c6][c64];
extern char * beta_sbox[c6][c64];
extern char *gamma_sbox[c6][c64];

/* stats.c */
void setup_stats(void);
void some_stats(char *what);

/* tests.c */
void test_stuff(void);

/* theta.c */
void setup_theta(void);
int theta_sim(int operation, int t_flag, int props, int carries);
uint16_t theta_calc(int i);

/* util.c */
void oops(char *fmt, ...);
int sbox_unquote(int c);
int sbox_invert(char *q, int i);
void urandom(int len, void *buf);
long slow_wire(char *fmt, long A);
int rotate_tribble_left(int t, int n);
void print6(char *prefix, uint t, char *suffix);
void print36(char *prefix, ulong w, char *suffix);

extern int show;		/* console view */
extern int tracing;		/* diagnostic */
extern int use_ram;		/* diagnostic */
extern const long pole, two36, mask36, comp36;

#define trace if (tracing) printf
